"""Test for data conversion functions."""

import json
from math import ceil
from operator import itemgetter
import os
import sys
import tempfile
import time
import unittest

import networkx as nx
import numpy as np
import pytest
import scipy.sparse as sp

from qci_client.data_converter import (
    compute_results_step_len,
    data_to_json,
    load_json_file,
)


@pytest.mark.offline
class TestDataConverter(unittest.TestCase):
    """Test suite for data conversion."""

    def test_load_json_file(self):
        """Test loading a JSON file."""
        test_dict = {
            "str": "hello",
            "int": 56,
            "float": 8.6,
            "list": [1, 2, 3],
            "nestlist": [{"i": 1, "j": 1, "val": 0.1}, {"i": 1, "j": 1, "val": 0.1}],
        }

        # Serializing json
        json_object = json.dumps(test_dict)
        # write to tempdir
        test_dir = tempfile.mkdtemp()
        file_name = str(os.path.join(test_dir, "test.json"))
        with open(file_name, "w", encoding="utf-8") as file_out:
            file_out.write(json_object)
        dict_load = load_json_file(file_name=file_name)
        self.assertDictEqual(test_dict, dict_load)


@pytest.mark.offline
class TestDataToJson(unittest.TestCase):
    """Test suite for data conversion to JSON."""

    #def test_file_type_assert(self):
    #    """Test file generation for bad file type."""
    #    # test file_type assertion
    #    with self.assertRaises(AssertionError):
    #        data_to_json(data=[])

    #def test_filename_blank(self):
    #    """Test file generation for missing file name."""
    #    # test file_name change if blank
    #    name_change = data_to_json(data=[])
    #    self.assertEqual(name_change["file_name"], "rhs.json")

    #def test_file_type_and_name(self):
    #    """Test file generation for file type and name."""
    #    file_name_check = data_to_json(
    #        data=[]
    #    )
    #    #self.assertEqual(file_name_check["file_type"], "rhs")
    #    self.assertEqual(file_name_check["file_name"], "other_name.json")

    def test_graph_file_body(self):
        """Test file generation for graph data."""
        graph = nx.Graph()
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        #graph_dict_check = {
        #    "file_type": "graph",
        #    "file_name": "graph.json",
        #    "directed": False,
        #    "multigraph": False,
        #    "graph": {},
        #    "nodes": [{"id": 1}, {"id": 2}, {"id": 3}],
        #    "links": [{"source": 1, "target": 2}, {"source": 1, "target": 3}],
        #}
        graph_dict_check = {
            "file_name": "graph.json",
            "file_config": {
                "graph": {
                    "directed": False,
                    "multigraph": False,
                    "graph": {},
                    "nodes": [{"id": 1}, {"id": 2}, {"id": 3}],
                    "links": [{"source": 1, "target": 2}, {"source": 1, "target": 3}],
                },
            },
        }

        #graph_body = data_to_json(data=graph)
        #self.assertDictEqual(graph_dict_check, graph_body)

        #with self.assertRaises(AssertionError) as context:
        #    data_to_json(data=np.array([]))

        #self.assertEqual(
        #    str(context.exception), "'graph' file_type must be type networkx.Graph"
        #)

    @pytest.mark.timing
    def test_large_data_conversion(self):
        """Test file generation for large data is sufficiently fast."""
        large_qubo = np.random.normal(size=(3000, 3000))
        large_qubo = large_qubo + large_qubo.T
        start = time.perf_counter()
        data_to_json(data=large_qubo)
        end = time.perf_counter()
        conversion_time = end - start

        self.assertTrue(
            conversion_time < 5,
            msg=f"Matrix conversion to JSON took too long: 5s <= {conversion_time}s.",
        )

    #def test_type_not_graph_check(self):
    #    """Test file generation for mismatched data and problem types."""
    #    with self.assertRaises(AssertionError) as context:
    #        graph = nx.Graph()
    #        data_to_json(data=graph)

    #    self.assertEqual(
    #        str(context.exception),
    #        "file_types ['rhs', 'objective', 'qubo', 'constraints', 'constraint_penalties', 'hamiltonian'] do not support networkx.Graph type",
    #    )

    def test_rhs_file_body(self):  # pylint: disable=too-many-locals
        """Test file generation for right-hand-side data."""
        rhs_list = [1, 2, 3]
        rhs_np = np.array([1, 2, 3])
        rhs_np_long = np.array([[1], [2], [3]])
        rhs_sp = sp.csr_matrix(rhs_np)
        rhs_sp_long = sp.csr_matrix(rhs_np_long)
        rhs_body_check = {
            "file_name": "rhs.json",
            "file_config": {
                "rhs": {
                    "num_constraints": 3,
                    "data": [1, 2, 3],
                },
            },
        }

        #commenting out, not currently supported
        #rhs_list_body = data_to_json(data=rhs_list)
        #self.assertDictEqual(rhs_body_check, rhs_list_body)
        #rhs_np_body = data_to_json()
        #self.assertDictEqual(rhs_body_check, rhs_np_body)
        #rhs_sp_body = data_to_json(file_type="rhs")
        #self.assertDictEqual(rhs_body_check, rhs_sp_body)
        #rhs_np_long_body = data_to_json(file_type="rhs")
        #self.assertDictEqual(rhs_body_check, rhs_np_long_body)
        #rhs_sp_long_body = data_to_json(file_type="rhs")
        #self.assertDictEqual(rhs_body_check, rhs_sp_long_body)

        constraint_penalties_list = [1, 2, 3]
        constraint_penalties_np = np.array([1, 2, 3])
        constraint_penalties_np_long = np.array([[1], [2], [3]])
        constraint_penalties_sp = sp.csr_matrix(constraint_penalties_np)
        constraint_penalties_sp_long = sp.csr_matrix(constraint_penalties_np_long)

        #constraint_penalties_body_check = {
        #    "file_type": "constraint_penalties",
        #    "file_name": "constraint_penalties.json",
        #    "num_constraints": 3,
        #    "data": [1, 2, 3],
        #}
        constraint_penalties_body_check = {
            "file_name": "constraint_penalties.json",
            "file_config": {
                "constraint_penalties": {
                    "num_constraints": 3,
                    "data": [1, 2, 3],
                }
            }
        }

        #constraint_penalties_list_body = data_to_json(
        #    data=constraint_penalties_list
        #)
        #self.assertDictEqual(
        #    constraint_penalties_body_check, constraint_penalties_list_body
        #)
        #constraint_penalties_np_body = data_to_json(
        #    data=constraint_penalties_np
        #)
        #self.assertDictEqual(
        #    constraint_penalties_body_check, constraint_penalties_np_body
        #)
        #constraint_penalties_sp_body = data_to_json(
        #    data=constraint_penalties_sp
        #)
        #self.assertDictEqual(
        #    constraint_penalties_sp_body
        #)
        #constraint_penalties_np_long_body = data_to_json(
        #    data=constraint_penalties_np_long
        #)
        #self.assertDictEqual(
        #    constraint_penalties_body_check, constraint_penalties_np_long_body
        #)
        #constraint_penalties_sp_long_body = data_to_json(
        #    data=constraint_penalties_sp_long
        #)
        #self.assertDictEqual(
        #    constraint_penalties_body_check, constraint_penalties_sp_long_body
        #)

    #def test_assert_types_objective_matrix(self):
    #    """Test file generation for improperly formatted qubo data."""
    #    with self.assertRaises(AssertionError) as context:
    #        data_to_json(data=[[1, -1], [-1, 1]])

    #    self.assertEqual(
    #        str(context.exception),
    #        "file_types = ['qubo', 'objective', 'constraints', 'hamiltonian'] only support types np.ndarray and scipy.sparse.spmatrix",
    #    )

    def test_qubo_hamiltonian_constraints_objective_file_body(self):
        """
        Test file generation for a qubo and hamiltonian objectives with constraints.
        """
        # will be used for both qubo and objective since same shape
        q_obj_np = np.array([[-1, 1], [1, -1]])
        q_obj_sp = sp.csr_matrix(q_obj_np)
        # is in i,j sorted order to allow for exact match of lists
        q_obj_data = [
            {"i": 0, "j": 0, "val": -1.0},
            {"i": 0, "j": 1, "val": 1.0},
            {"i": 1, "j": 0, "val": 1.0},
            {"i": 1, "j": 1, "val": -1.0},
        ]

        # using for hamiltonian and constraints
        ham_cons_np = np.array([[-1, 1, 1], [1, -1, 1]])
        ham_cons_sp = sp.csr_matrix(ham_cons_np)
        # is i, j sorted order to allow for exact match of lists
        ham_cons_data = [
            {"i": 0, "j": 0, "val": -1.0},
            {"i": 0, "j": 1, "val": 1.0},
            {"i": 0, "j": 2, "val": 1.0},
            {"i": 1, "j": 0, "val": 1.0},
            {"i": 1, "j": 1, "val": -1.0},
            {"i": 1, "j": 2, "val": 1.0},
        ]

        json_template = {
            "file_name": "placeholder",
            "file_config": {
            }
        }

        # start from using fewest fields to most so can use update on same json_template
        # qubo
        file_type = "qubo"
        json_template.update(
            {
                "file_name": file_type + ".json",
                "file_config": {
                    file_type: {
                        "data": q_obj_data,
                        "num_variables": 2,
                    }
                }
            }
        )
        #qubo_np_body = data_to_json(data=q_obj_np)
        #qubo_np_body["file_config"][file_type]["data"] = sorted(qubo_np_body["file_config"][file_type]["data"], key=itemgetter("i", "j"))
        #self.assertDictEqual(json_template, qubo_np_body)
        # objective
        file_type = "objective"
        #json_template.update({"file_type": file_type, "file_name": file_type + ".json"})
        #objective_sp_body = data_to_json(data=q_obj_sp)
        #objective_sp_body["file_config"][file_type]["data"] = sorted(
        #    objective_sp_body["file_config"][file_type]["data"], key=itemgetter("i", "j")
        #)
        #self.assertDictEqual(json_template, objective_sp_body)
        # hamiltonian
        file_type = "hamiltonian"
        # don't have to update num_variables becaues is the same as was used for qubo and objective
        json_template.update(
            {
                "file_name": file_type + ".json",
                "file_config": {
                    file_type: {
                        "data": ham_cons_data,
                        "num_variables": 2,
                    }
                }
            }
        )
        #ham_np_body = data_to_json(data=ham_cons_np)
        #ham_np_body["file_config"][file_type]["data"] = sorted(ham_np_body["file_config"][file_type]["data"], key=itemgetter("i", "j"))
        #print(ham_np_body)
        #print(json_template)
        #self.assertDictEqual(json_template, ham_np_body)
        # objective
        file_type = "constraints"
        json_template.update(
            {
                "file_name": file_type + ".json",
                "file_config": {
                    file_type: {
                        "num_variables": 3,
                        "num_constraints": 2,
                    }
                }
            }
        )
        #constraints_sp_body = data_to_json(data=ham_cons_sp)
        #constraints_sp_body["file_config"][file_type]["data"] = sorted(
        #    constraints_sp_body["file_config"][file_type]["data"], key=itemgetter("i", "j")
        #)
        #print(constraints_sp_body)
        #print(json_template)
        #self.assertDictEqual(json_template, constraints_sp_body)

    def test_small_graph(self):
        """Test file generation for a small graph problem."""
        graph = nx.Graph()
        edge_wt = 1.0
        graph.add_nodes_from(range(7))

        for node in graph.nodes:
            if node in [0, 1, 2, 3]:
                graph.nodes[node].update({"bipartite": 0})
            else:
                graph.nodes[node].update({"bipartite": 1})

        graph.add_edges_from(
            [
                (0, 4, {"weight": edge_wt}),
                (0, 6, {"weight": edge_wt}),
                (1, 4, {"weight": edge_wt}),
                (2, 5, {"weight": edge_wt}),
                (3, 5, {"weight": edge_wt}),
            ]
        )

        graph_dict_check = {
            "file_type": "graph",
            "file_name": "graph.json",
            "directed": False,
            "multigraph": False,
            "graph": {},
            "nodes": list(graph.nodes(data=True)),
            "links": list(graph.edges(data=True)),
        }

        # get the first element (only one), and also remove the part_num (second tuple element)
        compress = False
        legacy = False



"""
Utilities for Python client for files REST API for optimization service.

Copyright 2023, Quantum Computing Incorporated
"""

import gzip
from io import BytesIO
import json
from math import floor
import sys
from typing import Generator

#from optimization_files_api_py_client import enum, types
from qci_client import enum, types

# We want to limit the memory size of each uploaded chunk to be safely below the max of
# 15 * MebiByte (~15MB). See https://git.qci-dev.com/qci-dev/optimization-files-api/.
MEMORY_MAX: int = 8 * 1000000  # 8MB


def get_file_type(file: dict) -> enum.FileType:
    """Get file type from file."""
    file_config = file["file_config"]
    file_config_entries = list(file_config.keys())

    if len(file_config_entries) != 1:
        raise ValueError("file must specify exactly one file type under file_config")

    return enum.FileType(file_config_entries[0])


def get_post_request_body(  # pylint: disable=too-many-branches,too-many-return-statements
    *, file: dict
) -> types.MetadataPostRequestBody:
    """Format metadata body."""
    file_type = get_file_type(file=file)
    file_config = file["file_config"][file_type.value]
    optional_fields = {}

    if "file_name" in file:
        optional_fields["file_name"] = file["file_name"]

    if file_type == enum.FileType.CONSTRAINTS:
        return types.InputMetadataPostRequestBody(
            **optional_fields,
            file_config=types.ConstraintsMetadataConfig(
                constraints=types.ConstraintsMetadata(
                    num_constraints=file_config["num_constraints"],
                    num_variables=file_config["num_variables"],
                )
            ),
        )

    if file_type == enum.FileType.GRAPH:
        if "directed" in file_config:
            optional_fields["directed"] = file_config["directed"]

        if "multigraph" in file_config:
            optional_fields["multigraph"] = file_config["multigraph"]

        if "graph" in file_config:
            optional_fields["graph"] = file_config["graph"]

        return types.InputMetadataPostRequestBody(
            **optional_fields,
            file_config=types.GraphMetadataConfig(
                graph=types.GraphMetadata(
                    **optional_fields,
                    num_edges=file_config["num_edges"],
                    num_nodes=file_config["num_nodes"],
                )
            ),
        )

    if file_type == enum.FileType.HAMILTONIAN:
        return types.InputMetadataPostRequestBody(
            **optional_fields,
            file_config=types.HamiltonianMetadataConfig(
                hamiltonian=types.HamiltonianMetadata(
                    num_variables=file_config["num_variables"],
                )
            ),
        )

    if file_type == enum.FileType.OBJECTIVE:
        return types.InputMetadataPostRequestBody(
            **optional_fields,
            file_config=types.ObjectiveMetadataConfig(
                objective=types.ObjectiveMetadata(
                    num_variables=file_config["num_variables"],
                )
            ),
        )

    if file_type == enum.FileType.QUBO:
        return types.InputMetadataPostRequestBody(
            **optional_fields,
            file_config=types.QuboMetadataConfig(
                qubo=types.QuboMetadata(
                    num_variables=file_config["num_variables"],
                )
            ),
        )

    if file_type == enum.FileType.GP_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.GpResultsMetadataConfig(
                graph_partitioning_results=types.GpResultsMetadata()
            ),
        )

    if file_type == enum.FileType.IHO_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.IhoResultsMetadataConfig(
                ising_hamiltonian_optimization_results=types.IhoResultsMetadata()
            ),
        )

    if file_type == enum.FileType.NQHO_CONTINUOUS_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.NqhoContinuousResultsMetadataConfig(
                normalized_qudit_hamiltonian_optimization_continuous_results=types.NqhoContinuousResultsMetadata()  # pylint: disable=line-too-long
            ),
        )

    if file_type == enum.FileType.NQHO_INTEGER_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.NqhoIntegerResultsMetadataConfig(
                normalized_qudit_hamiltonian_optimization_integer_results=types.NqhoIntegerResultsMetadata()  # pylint: disable=line-too-long
            ),
        )

    if file_type == enum.FileType.QLCBO_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.QlcboResultsMetadataConfig(
                quadratic_linearly_constrained_binary_optimization_results=types.QlcboResultsMetadata()  # pylint: disable=line-too-long
            ),
        )

    if file_type == enum.FileType.QUBO_RESULTS:
        return types.ResultsMetadataPostRequestBody(
            **optional_fields,
            user_id=file["user_id"],
            organization_id=file["organization_id"],
            file_config=types.QuboResultsMetadataConfig(
                quadratic_unconstrained_binary_optimization_results=types.QuboResultsMetadata()  # pylint: disable=line-too-long
            ),
        )

    raise ValueError(f"unsupported file type: '{file_type.value}'")


def get_patch_request_body(  # pylint: disable=too-many-return-statements
    *, file: dict
) -> types.PartPatchRequestBody:
    """Format part body."""
    file_type = get_file_type(file=file)
    file_config = file["file_config"][file_type.value]

    if file_type == enum.FileType.CONSTRAINTS:
        return types.PartPatchRequestBody(
            file_config=types.ConstraintsPartConfig(
                constraints=types.ConstraintsPart(data=file_config["data"])
            ),
        )

    if file_type == enum.FileType.GRAPH:
        return types.PartPatchRequestBody(
            file_config=types.GraphPartConfig(
                graph=types.GraphPart(
                    links=file_config["links"],
                    nodes=file_config["nodes"],
                )
            ),
        )

    if file_type == enum.FileType.HAMILTONIAN:
        return types.PartPatchRequestBody(
            file_config=types.HamiltonianPartConfig(
                hamiltonian=types.HamiltonianPart(data=file_config["data"])
            ),
        )

    if file_type == enum.FileType.OBJECTIVE:
        return types.PartPatchRequestBody(
            file_config=types.ObjectivePartConfig(
                objective=types.ObjectivePart(data=file_config["data"])
            ),
        )

    if file_type == enum.FileType.QUBO:
        return types.PartPatchRequestBody(
            file_config=types.QuboPartConfig(
                qubo=types.QuboPart(data=file_config["data"])
            ),
        )

    if file_type == enum.FileType.GP_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.GpResultsPartConfig(
                graph_partitioning_results=types.GpResultsPart(
                    balances=file_config["balances"],
                    counts=file_config["counts"],
                    cut_sizes=file_config["cut_sizes"],
                    energies=file_config["energies"],
                    feasibilities=file_config["feasibilities"],
                    partitions=file_config["partitions"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    if file_type == enum.FileType.IHO_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.IhoResultsPartConfig(
                ising_hamiltonian_optimization_results=types.IhoResultsPart(
                    counts=file_config["counts"],
                    energies=file_config["energies"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    if file_type == enum.FileType.NQHO_CONTINUOUS_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.NqhoContinuousResultsPartConfig(
                normalized_qudit_hamiltonian_optimization_continuous_results=types.NqhoContinuousResultsPart(  # pylint: disable=line-too-long
                    counts=file_config["counts"],
                    energies=file_config["energies"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    if file_type == enum.FileType.NQHO_INTEGER_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.NqhoIntegerResultsPartConfig(
                normalized_qudit_hamiltonian_optimization_integer_results=types.NqhoIntegerResultsPart(  # pylint: disable=line-too-long
                    counts=file_config["counts"],
                    energies=file_config["energies"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    if file_type == enum.FileType.QLCBO_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.QlcboResultsPartConfig(
                quadratic_linearly_constrained_binary_optimization_results=types.QlcboResultsPart(  # pylint: disable=line-too-long
                    counts=file_config["counts"],
                    energies=file_config["energies"],
                    feasibilities=file_config["feasibilities"],
                    objective_values=file_config["objective_values"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    if file_type == enum.FileType.QUBO_RESULTS:
        return types.PartPatchRequestBody(
            file_config=types.QuboResultsPartConfig(
                quadratic_unconstrained_binary_optimization_results=types.QuboResultsPart(  # pylint: disable=line-too-long
                    counts=file_config["counts"],
                    energies=file_config["energies"],
                    solutions=file_config["solutions"],
                )
            ),
        )

    raise ValueError(f"unsupported file type: '{file_type.value}'")


def zip_payload(*, payload: types.PartPatchRequestBody) -> bytes:
    """
    :param payload: str - json contents of file to be zipped

    :return: zipped request_body
    """
    with BytesIO() as fileobj:
        with gzip.GzipFile(fileobj=fileobj, mode="w", compresslevel=6) as file:
            file.write(json.dumps(payload).encode("utf-8"))
        return fileobj.getvalue()


def file_part_generator(*, file: dict, compress: bool) -> Generator:
    """
    Break file-to-upload's data dictionary into chunks, formatting correctly with each
    returned chunk.

    :param file: file to break up into file parts
    :param compress: whether or not file parts are to be compressed

    :return: generator of (part_body, part_number) tuples
    """
    if compress:
        # The user has chosen to compress their files for upload we want a large
        # chunksize to try to maximize compression for each of the chunks.
        # Prior to merged stack, this value was 200000, which was too big in unit tests.
        data_chunk_size_max = 20000
    else:
        # We are using the multipart upload as a validated sharding system that is
        # similar to Mongo GridFS. Mongo recommends 256KB for that system, this value
        # keeps uploaded chunks below this value. After some testing, we decided to
        # limit this to chunks of 10000 elements for performance reasons.
        data_chunk_size_max = 10000

    file_type = enum.get_file_type(file=file)
    file_config = file["file_config"][file_type.value]

    if file_type in enum.JOB_INPUTS_MATRIX_FILE_TYPES:
        return _data_generator(
            file_type=file_type,
            file_config=file_config,
            step_length=data_chunk_size_max,
        )

    if file_type == enum.FileType.GRAPH:
        return _graph_generator(
            file_type=file_type,
            file_config=file_config,
            step_length=data_chunk_size_max,
        )

    # For results data, the n^2 sized data is in the solutions field so chunk it up.
    if file_type in enum.JOB_RESULTS_FILE_TYPES:
        return _results_generator(
            file_type=file_type,
            file_config=file_config,
            step_length=_compute_results_step_len(file_config["solutions"][0]),
        )

    raise ValueError(f"unhandled file_type: {file_type.value}")


def _get_size(obj, seen=None) -> int:
    """
    Recursively finds size of objects

    :param obj: data object to recursively compute size of
    :param seen: takes a set and is used in the recursive step only to record whether an
        object has been counted yet.

    :return int:
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum(_get_size(v, seen) for v in obj.values())
        size += sum(_get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, "__dict__"):
        size += _get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(_get_size(i, seen) for i in obj)
    return size


def _get_soln_size(soln):
    # Check whether first entry is a graph node/class assignment, e.g.,
    # {'id': 4, 'class': 2}.
    if isinstance(soln[0], dict):
        return _get_size(soln)

    return sys.getsizeof(soln[0]) * len(soln)


def _compute_results_step_len(data: list) -> int:
    """
    Compute the step length for "chunking" the provided data.

    Args:
        data: A list of data

    Returns:
        The step length for "chunking" the data
    """
    # total mem size of soln vector
    soln_mem = _get_soln_size(data)
    # num_vars * step_len < 30k => step_len < 30k/num_vars
    chunk_ratio = MEMORY_MAX / soln_mem
    step_len = max(floor(chunk_ratio), 1)

    return step_len


def _data_generator(
    *, file_type: enum.FileType, file_config: dict, step_length: int
) -> Generator:
    # data may be empty, so use max against 1.
    for part_number, i in enumerate(
        range(0, max(1, len(file_config["data"])), step_length)
    ):
        chunk = {
            "file_config": {
                file_type.value: {
                    "data": file_config["data"][i : i + step_length],
                }
            }
        }

        yield chunk, part_number + 1  #  content endpoint has 1-based uploads


def _graph_generator(
    *, file_type: enum.FileType, file_config: dict, step_length: int
) -> Generator:
    # links and nodes may both be empty, so use max against 1.
    for part_number, i in enumerate(
        range(
            0,
            max(1, len(file_config["links"]), len(file_config["nodes"])),
            step_length,
        )
    ):
        chunk = {
            "file_config": {
                file_type.value: {
                    "links": file_config["links"][i : i + step_length],
                    "nodes": file_config["nodes"][i : i + step_length],
                }
            }
        }

        yield chunk, part_number + 1  #  content endpoint has 1-based uploads


def _results_generator(
    *, file_type: enum.FileType, file_config: dict, step_length: int
) -> Generator:
    for part_number, i in enumerate(
        range(0, max(1, len(file_config["solutions"])), step_length)
    ):
        chunk = {"file_config": {file_type.value: {}}}

        for key, value in file_config.items():
            chunk["file_config"][file_type.value][key] = value[i : i + step_length]

        yield chunk, part_number + 1  #  content endpoint has 1-based uploads



"""Test qci-client utilities."""

from concurrent.futures import Future, ThreadPoolExecutor
import time
import unittest

import pytest

from qci_client.utils import _buffered_generator

MAX_WORKERS = 8
NUM_TASKS = 10


def _short_task(*args, **kwargs) -> tuple:
    """Simulate short 0.1s task, with pass-through args and kwargs."""

    time.sleep(0.1)

    return args, kwargs


def _medium_task(*args, **kwargs) -> tuple:
    """Simulate medium 1s task, with pass-through args and kwargs."""

    time.sleep(1)

    return args, kwargs


def _long_task_timeout(*args, **kwargs) -> tuple:
    """Simulate long 5s task that may time out, with pass-through args and kwargs."""

    for _ in range(5):
        time.sleep(1)

        if args[0] == int(NUM_TASKS / 2):
            raise TimeoutError("This long task timed out.")

    return args, kwargs


@pytest.mark.offline
class TestBufferedGeneration(unittest.TestCase):
    """Test various usages of utils._buffered_generator."""

    def test__buffered_generator_typical_usage(self):
        """Test buffered generator using typical usage pattern."""
        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:

            def _short_task_wrapper(*args, **kwargs) -> Future:
                return executor.submit(_short_task, *args, **kwargs)

            args_kwargs_generator = (
                (
                    ("fixed arg", idx),
                    {"key_fixed": "value_fixed", f"key_{idx}": f"value_{idx}"},
                )
                for idx in range(NUM_TASKS)
            )
            generator = _buffered_generator(
                buffer_length=MAX_WORKERS,
                task=_short_task_wrapper,
                args_kwargs_generator=args_kwargs_generator,
            )

            idx = 0
            for generated_value in generator:
                self.assertEqual(
                    generated_value,
                    (
                        ("fixed arg", idx),
                        {f"key_{idx}": f"value_{idx}", "key_fixed": "value_fixed"},
                    ),
                )
                idx += 1

            generator.close()  # Should not raise.
            self.assertEqual(idx, NUM_TASKS)

    def test__buffered_generator_internal_exception(self):
        """Test buffered generator with exception inside generator."""
        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:

            def _long_task_timeout_wrapper(*args, **kwargs) -> Future:
                return executor.submit(_long_task_timeout, *args, **kwargs)

            args_kwargs_generator = (
                (
                    (idx,),
                    {},
                )
                for idx in range(NUM_TASKS)
            )
            generator = _buffered_generator(
                buffer_length=MAX_WORKERS,
                task=_long_task_timeout_wrapper,
                args_kwargs_generator=args_kwargs_generator,
            )

            # An exception should occur in one task.
            with self.assertRaises(TimeoutError):
                idx = 0
                for generated_value in generator:
                    self.assertEqual(generated_value, ((idx,), {}))
                    idx += 1

            generator.close()  # Should not raise.

    def test__buffered_generator_external_exception(self):
        """Test buffered generator with exception outside generator (during yield)."""
        with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:

            def _medium_task_wrapped(*args, **kwargs) -> Future:
                return executor.submit(_medium_task, *args, **kwargs)

            args_kwargs_generator = (
                (
                    (idx,),
                    {},
                )
                for idx in range(NUM_TASKS)
            )
            generator = _buffered_generator(
                buffer_length=MAX_WORKERS,
                task=_medium_task_wrapped,
                args_kwargs_generator=args_kwargs_generator,
            )

            idx = 0
            for generated_value in generator:
                self.assertEqual(generated_value, ((idx,), {}))

                if idx == int(NUM_TASKS / 2):
                    # Proceed as if an exception occurred in yield.
                    with self.assertRaises(StopIteration):
                        # Cancel tasks and close generator.
                        generator.throw(Exception)

                idx += 1

            generator.close()  # Should not raise.



"""Utilities for qci-client."""

import concurrent.futures
from typing import Callable, Generator, Optional, Union


def _buffered_generator(
    *,
    buffer_length: int,
    task: Callable,
    args_kwargs_generator: Generator,
    timeout: Optional[Union[int, float]] = 0.01,
) -> Generator:
    """
    Generator that maps concurrent tasks over iterables using a buffer to limit memory
    pressure.

    :param buffer_length: Length of FIFO buffer, often equal to a thread/process count.
    :param task: A future-returning function to execute concurrently.
    :param args_kwargs_generator: Generator of (args, kwargs) tuple to pass to each
        task, where args is a tuple and kwargs is a dictionary.
    :param timeout: Timeout to wait for future results to resolve before looping, [s].

    :return: Generator using a concurrently-filled FIFO buffer.

    Notes:
    - Caller is expected to enforce any resource limits affecting choice of arguments.
    - Caller is expected to throw any Exception during yield back to generator to cancel
      any pending tasks and close generator gracefully (via return raising StopIteration
      in caller).
    """

    def async_generator() -> Generator:  # pylint disable=too-many-branches
        """Generates results of a task computed concurrently via a FIFO buffer."""

        pending_results = []  # FIFO buffer.
        result_ready = False
        has_input = True

        # Outer loop is cancelled from caller by caller throwing an exception back to
        # generator from within the yield.
        while True:
            if pending_results:
                try:
                    # Check if oldest task is done. Yielding via timeout.
                    result = pending_results[0].result(timeout=timeout)
                except concurrent.futures.TimeoutError:
                    # Let loop iterate before another result check.
                    pass
                except Exception:  # Includes CancelledError.
                    # Cancel any/all tasks before bubbling up exception.
                    for pending_result in pending_results:
                        pending_result.cancel()
                    raise
                else:
                    # Remove oldest task that resolved of front of list.
                    pending_results.pop(0)
                    # Yield result after refilling FIFO buffer.
                    result_ready = True

            # Submit tasks into FIFO buffer for remaining input until max_worker
            # tasks have been added.
            while has_input and len(pending_results) < buffer_length:
                try:
                    args_generated, kwargs_generated = next(args_kwargs_generator)
                    pending_results.append(task(*args_generated, **kwargs_generated))
                except StopIteration:
                    # Generated input ran out.
                    has_input = False

            # Yield generated result, if ready.
            if result_ready:
                try:
                    yield result
                except Exception:  # pylint: disable=broad-exception-caught
                    # Cancel any/all tasks before returning.
                    for pending_result in pending_results:
                        pending_result.cancel()
                    return

                result_ready = False

            # If there are no pending results, then generator is finished.
            if not pending_results:
                return

    return async_generator()



"""Test suite for jobs run on a remote backend."""

from typing import Optional

import numpy as np

from qci_client import (
    JobStatus,
    data_to_json,
)
from tests.test_qci_client import TestQciClientFiles


class TestQciClientRemoteJobs(TestQciClientFiles):
    """Collection of tests for remote jobs."""

    def run_end_end(
        self, job_type: str, job_body: dict, results_key: Optional[list] = None
    ):
        """
        Utility function for testing end to end pipeline.
        :param job_type: one of _supported_job_types
        :param job_body: a validate job request
        Testing each in series:
            - submit_job
            - get_job_status
            - get_job_response
            - get_file
        """

        result_keys = self.result_keys if results_key is None else results_key

        job_id = self.q1.submit_job(job_body=job_body, job_type=job_type)

        self.assertIn("job_id", job_id)
        self.assertIsInstance(job_id["job_id"], str)

        status = self.q1.get_job_status(job_id=job_id["job_id"])
        self.assertIn("status", status)
        self.assertIn(status["status"], self.all_statuses)

        while status["status"] not in self.final_status:
            status = self.q1.get_job_status(job_id=job_id["job_id"])

        self.assertEqual(JobStatus.COMPLETED, status["status"])
        response = self.q1.get_job_response(job_id=job_id["job_id"], job_type=job_type)

        # test job info has appropriate keys
        self.assertEqual(self.job_info, set(response.keys()))

        is_results_file = True

        #result = self.q1.get_file(file_id=response["results"]["file_id"], is_results_file=is_results_file)

        #result = list(result)
        #self.assertTrue(all(i in result[0] for i in result_keys))


    def process_job_check(self, job_type, job_body):
        """Utility function for checking job types."""
        process_key = ["job_info", "results"]
        job_output = self.q1.process_job(
            job_type=job_type, job_body=job_body, wait=True
        )
        self.assertTrue(all(key in process_key for key in list(job_output.keys())))

    def test_large_job(self):
        """Test large sample-qubo job."""
        large_qubo = np.random.normal(size=(6000, 6000))
        large_qubo = large_qubo + large_qubo.T
        large_qubo_dict = data_to_json(
            data=large_qubo
        )
        large_file_id = self.q1.upload_file(file=large_qubo_dict)
        print("LARGE FILE ID", large_file_id)
        large_job_body = {
            "job_name": "large_job",
            "qubo_file_id": large_file_id,
            "params": {"sampler_type": "eqc1", "n_samples": 2},
        }

        self.process_job_check(job_type="sample-qubo", job_body=large_job_body)

    def test_process_qubo(self):
        """Test that sample-qubo job process can be checked."""
        self.process_job_check(job_type="sample-qubo", job_body=self.qubo_job_body)

    def test_process_constraint(self):
        """Test that sample-constraint job process can be checked."""
        self.process_job_check(job_type="sample-constraint", job_body=self.objective_job_body)

    def test_graph_partitioning(self):
        """Test graph-partitioning job."""
        job_type = "graph-partitioning"

        self.run_end_end(job_type=job_type, job_body=self.graph_job_body)

    def test_sample_qubo(self):
        """Test sample-qubo job."""
        job_type = "sample-qubo"

        self.run_end_end(job_type=job_type, job_body=self.qubo_job_body)

    def test_sample_constraint(self):
        """Test sample-constraint job."""
        job_type = "sample-constraint"

        self.run_end_end(job_type=job_type, job_body=self.objective_job_body)

    def test_uni_cd(self):
        """Test unipartite-community-detection job."""
        job_type = "unipartite-community-detection"

        self.run_end_end(job_type=job_type, job_body=self.graph_job_body)

    def test_process_hamiltonian(self):
        """Test that sample-hamiltonian job process can be checked."""
        self.process_job_check(
            job_type="sample-hamiltonian", job_body=self.hamiltonian_job_body
        )

    def test_sample_hamiltonian(self):
        """Test sample-hamiltonian job."""
        job_type = "sample-hamiltonian"

        self.run_end_end(job_type=job_type, job_body=self.hamiltonian_job_body)



"""Base class for API clients."""

from dataclasses import dataclass, field
import functools
import os
from posixpath import join
import time
from typing import Callable, Dict, Optional

import requests
from requests.adapters import HTTPAdapter, Retry

# TODO: temporary - for QCIEN-1142 (remove after we stop using verify=False on requests
from requests.packages.urllib3.exceptions import (  # pylint: disable=import-error,no-member
    InsecureRequestWarning,
)

requests.packages.urllib3.disable_warnings(  # pylint: disable=no-member
    InsecureRequestWarning
)

# We are uploading files we want to retry when we receive certain error codes
RETRY_TOTAL = 7
BACKOFF_FACTOR = 2
STATUS_FORCELIST = [502, 503, 504]


@dataclass
class BaseApi:  # pylint: disable=too-many-instance-attributes
    """Base class for API clients."""

    # Hide sensistive info to prevent accidental logging when printing client objects.
    _bearer_info: dict = field(default_factory=dict, repr=False)
    api_token: Optional[str] = field(default=None, repr=False)
    set_bearer_token_on_init: bool = True

    # User-definable url's and endpoints
    url: Optional[str] = None
    authorize: Optional[str] = None

    # Default endpoints and url
    _authorize: str = "auth/v1/access-tokens"
    _download: str = "contents"
    _add_headers: Dict[str, str] = None

    # Constants
    _user_not_authorized: str = "user not authorized"

    # Request timeout in seconds for connection & read. None for infinite timeout.
    timeout: Optional[float] = 5 * 60.0

    debug: bool = False

    def __post_init__(self):
        self.api_token = self.api_token if self.api_token else os.getenv("QCI_TOKEN")
        assert (
            self.api_token is not None
        ), "QCI_TOKEN environment variable is empty. Specify api_token or add the necessary environment variable"

        self.url = self.url if self.url else os.getenv("QCI_API_URL")
        assert (
            self.url is not None
        ), "QCI_API_URL environment variable is empty. Specify url or add the necessary environment variable"
        # removing trailling / so can add paths simply
        self.url.rstrip("/")

        self.authorize = self.authorize if self.authorize else self._authorize

        self.session = requests.Session()
        retries = Retry(
            total=RETRY_TOTAL,
            backoff_factor=BACKOFF_FACTOR,
            status_forcelist=STATUS_FORCELIST,
        )
        self.session.mount("https://", HTTPAdapter(max_retries=retries))

        if self.set_bearer_token_on_init:
            self.set_bearer_token()

    @property
    def auth_url(self) -> str:
        """Return the URL used for authorization."""

        return join(self.url, self.authorize)

    @property
    def headers_without_token(self):
        """Headers without cached bearer token."""
        headers = {
            "Content-Type": "application/json",
            "Connection": "close",
        }

        if self.timeout is not None:
            # Provide client's request timeout to server (latency provides some buffer).
            headers.update({"X-Request-Timeout-Nano": str(int(10**9 * self.timeout))})

        # allow additional headers to be passed through
        if self._add_headers is not None:
            headers.update(self._add_headers)

        return headers

    @property
    def headers(self):
        """Headers with cached bearer token."""
        headers = self.headers_without_token
        headers["Authorization"] = f"Bearer {self._bearer_info.get('access_token', '')}"

        return headers

    @property
    def headers_without_connection_close(self):
        """Headers with cached bearer token, but without connection closing."""
        headers = self.headers
        headers.pop("Connection", None)

        return headers

    def _check_response_error(self, response: requests.Response) -> None:
        """
        Single place to update error check and message for API calls
        :param response: a response from any API call using the requests package
        """
        assert (
            response.status_code < 300
        ), f"Error: {response.text}. Received error code {response.status_code}"

    def get_bearer_token(self) -> requests.Response:
        """Request new bearer token. (Not cached here, see set_bearer_token.)"""
        payload = {"refresh_token": self.api_token}

        response = self.session.request(
            "POST",
            self.auth_url,
            json=payload,
            headers=self.headers_without_token,
            timeout=self.timeout,
            verify=False,
        )

        self._check_response_error(response)

        return response

    def set_bearer_token(self) -> None:
        """Set bearer token from request."""
        resp = self.get_bearer_token()
        self._bearer_info = resp.json()

    def is_bearer_token_expired(self) -> bool:
        """
        Is current time > 'expires' time.
        TODO: expires_in should be 'expires_at'
        """
        # TODO: this should eventually read 'expires_at'
        # adding 10 second buffer for expiration
        return int(time.time()) + 10 >= self._bearer_info.get("expires_in", 0)

    @staticmethod
    def refresh_token(func) -> Callable:
        """Return a wrapper function that can check an auth token."""

        @functools.wraps(func)
        def check_token(api, *args, **kwargs):
            # Because the decorated function is receiving 'self', we need to pass this
            # additional argument along in the 'api' arg.
            is_expired = api.is_bearer_token_expired()
            # expired, reset the token
            if is_expired:
                api.set_bearer_token()
                return func(api, *args, **kwargs)
            # still have time on the token, so just pass the wrapped func through
            return func(api, *args, **kwargs)

        return check_token



"""
Missing tests:
 - failures should be accounted for under the BaseApi
 - Each job should also be covered for process_job only 3/5 currently are.
"""

from contextlib import redirect_stdout
from datetime import datetime
import io
from math import ceil
from typing import Optional
import unittest
import unittest.mock

import networkx as nx
import numpy as np
import pytest
import requests
import scipy.sparse as sp

from qci_client import (
    JobStatus,
    QciClient,
    compute_results_step_len,
    data_to_json,
)


@pytest.mark.offline
class TestJobStatus(unittest.TestCase):
    """JobStatus-related test suite."""

    def test_jobstatus(self):
        """Test all job statuses."""
        job_status = JobStatus()
        self.assertEqual(job_status.QUEUED, "QUEUED")
        self.assertEqual(job_status.SUBMITTED, "SUBMITTED")
        self.assertEqual(job_status.RUNNING, "RUNNING")
        self.assertEqual(job_status.COMPLETED, "COMPLETED")
        self.assertEqual(job_status.ERRORED, "ERRORED")
        self.assertEqual(job_status.CANCELLED, "CANCELLED")


class TestQciClientFiles(unittest.TestCase):
    """Files-API-related test suite."""

    @classmethod
    def setUpClass(cls):
        cls.q1 = QciClient()
        cls.graph_dict_input = {
            "file_type": "graph",
            "file_name": "graph.json",
            "directed": False,
            "multigraph": False,
            "graph": {},
            "nodes": [{"id": 1}, {"id": 2}, {"id": 3}],
            "links": [{"source": 1, "target": 2}, {"source": 1, "target": 3}],
        }

        cls.q_dict_input = {
            "file_type": "qubo",
            "file_name": "qubo.json",
            "data": [
                {"i": 0, "j": 0, "val": -1.0},
                {"i": 0, "j": 1, "val": 1.0},
                {"i": 1, "j": 0, "val": 1.0},
                {"i": 1, "j": 1, "val": -1.0},
            ],
            "num_variables": 2,
        }

        cls.objective_dict_input = {
            "file_type": "objective",
            "file_name": "objective.json",
            "data": [
                {"i": 0, "j": 0, "val": -1.0},
                {"i": 0, "j": 1, "val": 1.0},
                {"i": 1, "j": 0, "val": 1.0},
                {"i": 1, "j": 1, "val": -1.0},
            ],
            "num_variables": 2,
        }

        cls.constraint_dict_input = {
            "file_type": "constraints",
            "file_name": "constraints.json",
            "data": [
                {"i": 0, "j": 0, "val": -1.0},
                {"i": 0, "j": 1, "val": 1.0},
                {"i": 1, "j": 0, "val": 1.0},
                {"i": 1, "j": 1, "val": -1.0},
            ],
            "num_variables": 2,
            "num_constraints": 2,
        }

        cls.rhs_dict_input = {
            "file_type": "rhs",
            "file_name": "rhs.json",
            "data": [1, 2],
            "num_constraints": 2,
        }

        cls.hamiltonian_dict_input = {
            "file_type": "hamiltonian",
            "file_name": "hamiltonian.json",
            "num_variables": 2,
            "data": [
                {"i": 0, "j": 0, "val": 1.0},
                {"i": 0, "j": 1, "val": 1.0},
                {"i": 1, "j": 0, "val": 2.0},
                {"i": 1, "j": 2, "val": 1.0},
            ],
        }

        cls.graph_file_id = cls.q1.upload_file(file=cls.graph_dict_input)["file_id"]
        cls.qubo_file_id = cls.q1.upload_file(file=cls.q_dict_input)["file_id"]
        cls.obj_file_id = cls.q1.upload_file(file=cls.objective_dict_input)["file_id"]
        cls.rhs_file_id = cls.q1.upload_file(file=cls.rhs_dict_input)["file_id"]
        cls.constraints_file_id = cls.q1.upload_file(file=cls.constraint_dict_input)[
            "file_id"
        ]
        cls.hamiltonian_file_id = cls.q1.upload_file(file=cls.hamiltonian_dict_input)[
            "file_id"
        ]

        cls.all_statuses = [
            JobStatus.QUEUED,
            JobStatus.SUBMITTED,
            JobStatus.RUNNING,
            JobStatus.COMPLETED,
            JobStatus.ERRORED,
            JobStatus.CANCELLED,
        ]

        cls.final_status = [JobStatus.COMPLETED, JobStatus.ERRORED, JobStatus.CANCELLED]

        cls.job_info = set(("job_id", "details", "results", "submission", "metrics"))
        cls.result_keys = ["samples", "energies", "file_name", "file_type"]

        cls.graph_job_body = {
            "job_name": "job_0",
            "graph_file_id": cls.graph_file_id,
            "params": {"sampler_type": "csample", "n_samples": 1}
        }

        cls.qubo_job_body = {
            "job_name": "job_0",
            "qubo_file_id": cls.qubo_file_id,
            "params": {"sampler_type": "csample", "n_samples": 1}
        }

        cls.objective_job_body = {
            "job_name": "job_0",
            "objective_file_id": cls.obj_file_id,
            "constraints_file_id": cls.constraints_file_id,
            "rhs_file_id": cls.rhs_file_id,
            "params": {"sampler_type": "csample", "n_samples": 1}
        }

        cls.hamiltonian_job_body = {
            "job_name": "job_0",
            "hamiltonian_file_id": cls.hamiltonian_file_id,
            "params": {"sampler_type": "eqc1", "n_samples": 2},
        }

    def run_end_end(
        self, job_type: str, job_body: dict, results_key: Optional[list] = None
    ):
        """
        Utility function for testing end to end pipeline.
        :param job_type: one of _supported_job_types
        :param job_body: a validate job request
        Testing each in series:
            - submit_job
            - get_job_status
            - get_job_response
            - get_file
        """

        result_keys = self.result_keys if results_key is None else results_key
        is_results_file = True

        job_id = self.q1.submit_job(job_body=job_body, job_type=job_type)

        self.assertIn("job_id", job_id)
        self.assertIsInstance(job_id["job_id"], str)

        status = self.q1.get_job_status(job_id=job_id["job_id"])
        self.assertIn("status", status)
        self.assertIn(status["status"], self.all_statuses)

        while status["status"] not in self.final_status:
            status = self.q1.get_job_status(job_id=job_id["job_id"])

        self.assertEqual(JobStatus.COMPLETED, status["status"])
        response = self.q1.get_job_response(job_id=job_id["job_id"], job_type=job_type)

        # test job info has appropriate keys
        self.assertEqual(self.job_info, set(response.keys()))

    def test_add_headers(self):
        """Test adding headers to client's requests."""
        q_h = QciClient(_add_headers={"X-Job-Id": "fake-job", "X-Job-Type": "job_type"})
        self.assertTrue(q_h.headers["X-Job-Id"], "fake-job")
        self.assertTrue(q_h.headers["X-Job-Type"], "job-type")


    def test_upload_file_error(self):
        """Test uploading improperly formatted file."""
        with self.assertRaises(KeyError):
            error_input = {
                "file_type": "graph",
                "file_name": "qubo.json",
                "data": [
                    {"i": 0, "j": 0, "val": -1.0},
                    {"i": 0, "j": 1, "val": 1.0},
                    {"i": 1, "j": 0, "val": 1.0},
                    {"i": 1, "j": 1, "val": -1.0},
                ],
                "num_variables": 2,
            }
            self.q1.upload_file(file=error_input)

    def test_upload_file(self):
        """Test uploading of various file types."""
        # create graph data
        graph = nx.Graph()
        graph.add_edge(1, 2)
        graph.add_edge(1, 3)
        # create rhs data
        rhs_list = [1, 2]
        # create objective data
        objective_np = np.array([[1, -1], [-1, 1]])
        cons_hamiltonian_sp = sp.csr_matrix(np.array([[1, 1, -1], [2, -1, 1]]))

        graph_dict_upload = self.q1.upload_file(file=self.graph_dict_input)
        self.assertIn("file_id", graph_dict_upload)
        self.assertIsInstance(graph_dict_upload["file_id"], str)

        qubo_dict_upload = self.q1.upload_file(file=self.q_dict_input)
        self.assertIn("file_id", qubo_dict_upload)
        self.assertIsInstance(qubo_dict_upload["file_id"], str)

        rhs_dict_upload = self.q1.upload_file(file=self.rhs_dict_input)
        self.assertIn("file_id", rhs_dict_upload)
        self.assertIsInstance(rhs_dict_upload["file_id"], str)

        # upload_file takes a dict
        rhs_dict = data_to_json(
            data=rhs_list
        )
        rhs_list_upload = self.q1.upload_file(file=rhs_dict)
        self.assertIn("file_id", rhs_list_upload)
        self.assertIsInstance(rhs_list_upload["file_id"], str)

        graph_dict = data_to_json(
            data=graph
        )
        g_nx_upload = self.q1.upload_file(file=graph_dict)
        self.assertIn("file_id", g_nx_upload)
        self.assertIsInstance(g_nx_upload["file_id"], str)

        objective_dict = data_to_json(
            data=objective_np
        )
        obj_np_upload = self.q1.upload_file(file=objective_dict)
        self.assertIn("file_id", obj_np_upload)
        self.assertIsInstance(obj_np_upload["file_id"], str)

        hamiltonian_dict = data_to_json(
            data=cons_hamiltonian_sp
        )
        ham_sp_upload = self.q1.upload_file(
            file=hamiltonian_dict
        )
        self.assertIn("file_id", ham_sp_upload)
        self.assertIsInstance(ham_sp_upload["file_id"], str)

        constraints_dict = data_to_json(
            data=cons_hamiltonian_sp
        )
        cons_sp_upload = self.q1.upload_file(
            file=constraints_dict
        )
        self.assertIn("file_id", cons_sp_upload)
        self.assertIsInstance(cons_sp_upload["file_id"], str)

    def test_validate_job_type(self):
        """Test validation of job type against whitelist."""
        test_job_type = "fake_job_type"
        expected_err_str = f"Provided job_type '{test_job_type}' is not one of {self.q1._supported_job_types}"  # pylint: disable=protected-access

        with self.assertRaises(AssertionError) as context:
            self.q1.validate_job_type(job_type=test_job_type)

        # TODO: consider asserting on exception type instead of specific text of assert msg
        # because aserting on the specific text of the assert msg makes this test fairly fragile
        self.assertEqual(
            str(context.exception),
            expected_err_str
        )

        try:
            self.q1.validate_job_type(job_type="sample-qubo")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.fail(
                f"validate_job_type failed with exception '{exc}' with supported job_type"
            )

    def test_build_job_body(self):
        """Test building of various jobs' request bodies."""
        # don't need to revalidate validate_job_type
        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(
                job_type="sample-qubo",
                qubo_file_id="qubo_file_id",
                objective_file_id="objective_file_id",
            )

        self.assertEqual(
            str(context.exception),
            "Only one of qubo_file_id, hamiltonian_file_id, objective_file_id, or graph_file_id can be specified",
        )

        job_template = {"job_name": "job_0", "job_tags": [], "params": {}}

        # Validate sample-constraint
        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(job_type="sample-constraint", qubo_file_id="q_fid")

        self.assertEqual(
            str(context.exception),
            "objective_file_id, constraints_file_id, and rhs_file_id must all be specified for job_type='sample-constraint'",
        )

        constraint_template = {
            "objective_file_id": "obj_fid",
            "constraints_file_id": "cons_fid",
            "rhs_file_id": "rhs_fid",
        }

        constraint_template.update(job_template)

        constraint_body = self.q1.build_job_body(
            job_type="sample-constraint",
            objective_file_id="obj_fid",
            constraints_file_id="cons_fid",
            rhs_file_id="rhs_fid",
        )

        self.assertDictEqual(constraint_body, constraint_template)

        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(job_type="sample-qubo", objective_file_id="obj_fid")

        self.assertEqual(
            str(context.exception),
            "objective_file_id, constraints_file_id, and rhs_file_id are not available for selected job_type",
        )

        # Validate sample-qubo
        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(
                job_type="sample-qubo", hamiltonian_file_id="obj_fid"
            )

        self.assertEqual(
            str(context.exception),
            "qubo_file_id must be specified for job_type='sample-qubo'",
        )

        qubo_template = {"qubo_file_id": "q_fid"}
        qubo_template.update(job_template)

        qubo_body = self.q1.build_job_body(job_type="sample-qubo", qubo_file_id="q_fid")

        self.assertDictEqual(qubo_template, qubo_body)

        # Validate sample-hamiltonian
        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(job_type="sample-hamiltonian", qubo_file_id="q_fid")

        self.assertEqual(
            str(context.exception),
            "hamiltonian_file_id must be specified for job_type='sample-hamiltonian'",
        )

        hamiltonian_template = {"hamiltonian_file_id": "h_fid"}
        hamiltonian_template.update(job_template)

        hamiltonian_body = self.q1.build_job_body(
            job_type="sample-hamiltonian", hamiltonian_file_id="h_fid"
        )

        self.assertDictEqual(hamiltonian_template, hamiltonian_body)

        # Validate graph problems
        with self.assertRaises(AssertionError) as context:
            self.q1.build_job_body(
                job_type="graph-partitioning", hamiltonian_file_id="obj_fid"
            )

        self.assertEqual(
            str(context.exception),
            "graph_file_id must be specified for the given job_type",
        )

        graph_template = {"graph_file_id": "g_fid"}
        graph_template.update(job_template)

        graph_body = self.q1.build_job_body(
            job_type="bipartite-community-detection", graph_file_id="g_fid"
        )

        self.assertDictEqual(graph_template, graph_body)

        # test other params
        name_body = self.q1.build_job_body(
            job_type="unipartite-community-detection",
            graph_file_id="g_fid",
            job_name="fake_job",
        )
        graph_template.update({"job_name": "fake_job", "job_tags": []})

        self.assertDictEqual(graph_template, name_body)

        tag_body = self.q1.build_job_body(
            job_type="unipartite-community-detection",
            graph_file_id="g_fid",
            job_tags=["tag1", "tag2"],
        )
        graph_template.update({"job_name": "job_0", "job_tags": ["tag1", "tag2"]})
        self.assertDictEqual(graph_template, tag_body)

        job_params_body = self.q1.build_job_body(
            job_type="unipartite-community-detection",
            graph_file_id="g_fid",
            job_params={"n_shots": 100, "device": "ionq"},
        )
        graph_template.update(job_template)
        graph_template.update({"params": {"n_shots": 100, "device": "ionq"}})
        self.assertDictEqual(graph_template, job_params_body)

    def test_print_job_log(self):
        """Test printing of job log messages."""
        string_io = io.StringIO()
        with redirect_stdout(string_io):
            self.q1.print_job_log(message="fake message")
        out = string_io.getvalue()
        date_now = str(datetime.now().strftime("%Y/%m/%d %H:%M:%S"))[0:10]
        self.assertTrue(out.startswith(f"fake message: {date_now}"))


class TestMultipartUpload(unittest.TestCase):
    """Multipart-file upload test suite."""

    @classmethod
    def setUpClass(cls):
        cls.q1 = QciClient()

    def test_multipart_results_upload_and_download(self):
        """Test uploading/downloading multipart qubo results file."""
        # BIG number of samples, each one num_vars in length
        num_vars = 1700
        num_samp = 300
        is_results_file = True
        samples = np.ones((num_samp, num_vars)).astype(int)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_qubo",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(resdata["samples"][0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        #resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), counts.shape[0])
        #self.assertEqual(len(test_res_whole["energies"]), energies.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), samples.shape[0])
        #self.assertEqual(len(test_res_whole["samples"][0]), samples.shape[1])

        #del_resp = self.q1.delete_file(resp["file_id"])
        #assert del_resp["num_deleted"] == 1

    def test_multipart_results_upload_and_download_hamiltonian(self):
        """Test uploading/downloading multipart hamiltonian results file."""
        # Tests float uploads, so we use Hamiltonian job type so the API can handle floats
        num_vars = 20000
        num_samp = 30
        samples = np.ones((num_samp, num_vars)).astype(float)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))
        is_results_file = True

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_hamiltonian",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            #"solution_type": "continuous",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(resdata["samples"][0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), len(test_vec_int))
        #self.assertEqual(len(test_res_whole["energies"]), test_vec.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), num_samples)
        #self.assertEqual(len(test_res_whole["samples"][0]), num_nodes)

        del_resp = self.q1.delete_file(resp["file_id"])
        assert del_resp["num_deleted"] == 1

    def test_multipart_results_upload_and_download_huge(self):
        """Test uploading/downloading very large multipart qubo results file."""
        # Now try something too large
        # BIG number of samples, each one num_vars in length
        num_vars = 100000
        num_samp = 300
        samples = np.ones((num_samp, num_vars)).astype(int)
        counts = np.ones((num_samp, 1)).astype(int)
        energies = np.ones((num_samp, 1))
        is_results_file = True

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_sample_qubo",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "counts": counts.flatten().tolist(),
            "energies": energies.flatten().tolist(),
            "samples": samples.tolist(),
        }

        step_len = compute_results_step_len(samples[0])
        expected_parts = ceil(num_samp / step_len)
        self.assertGreater(expected_parts, 1)

        resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), counts.shape[0])
        #self.assertEqual(len(test_res_whole["energies"]), energies.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), samples.shape[0])
        #self.assertEqual(len(test_res_whole["samples"][0]), samples.shape[1])


    def test_multipart_graph_partitioning_results_upload_and_download(self):
        """Test uploading/downloading multipart graph-partitioning results file."""
        num_samples = 10
        num_nodes = 10000
        test_vec = np.arange(num_samples)
        test_vec_int = np.arange(num_samples).astype(int)
        is_results_file = True

        samples = [
            [
                {"id": int(np.random.randint(0, 10)), "class": np.random.randint(0, 3)}
                for _ in range(num_nodes)
            ]  # pretend we have a graph with 100 nodes
            for _ in range(num_samples)  # and we asked for 1000 samples
        ]

        gp_results = {
            "file_name": "test-file.json",
            "file_type": "job_results_graph_partitioning",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "balance": test_vec.tolist(),
            "counts": test_vec.tolist(),
            "cut_size": test_vec_int.tolist(),
            "energies": test_vec.tolist(),
            "is_feasible": [True] * num_samples,
            "samples": samples,
        }

        step_len = compute_results_step_len(samples[0])
        expected_parts = ceil(num_samples / step_len)
        self.assertGreater(expected_parts, 1)

        resp = self.q1.upload_file(file=gp_results)

        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #del_resp = self.q1.delete_file(resp["file_id"])
        #assert del_resp["num_deleted"] == 1

    def test_multipart_community_detection_results_upload_and_download(self):
        """Test uploading/downloading multipart community-detection results file."""
        num_samples = 1000
        num_nodes = 50
        test_vec = np.arange(num_samples)
        test_vec_int = np.arange(num_samples).astype(int)
        is_results_file = True

        samples = [
            [
                {"id": int(np.random.randint(0, 10)), "class": np.random.randint(0, 3)}
                for _ in range(num_nodes)
            ]  # pretend we have a graph with 100 nodes
            for _ in range(num_samples)  # and we asked for 1000 samples
        ]

        resdata = {
            "file_name": "test-file.json",
            "file_type": "job_results_bipartite_community_detection",
            "organization_id": "5ddf5db3fed87d53b6bf392a",
            "username": "test_user",
            "modularity": test_vec.tolist(),
            "counts": test_vec.tolist(),
            "energies": test_vec.tolist(),
            "is_feasible": [True] * num_samples,
            "samples": samples,
        }

        step_len = compute_results_step_len(resdata["samples"][0])
        expected_parts = ceil(num_samples / step_len)
        self.assertGreater(expected_parts, 1)

        resp = self.q1.upload_file(file=resdata)
        #meta = self.q1.get_file_metadata(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(meta["num_parts"], expected_parts)

        #test_res_whole = self.q1.get_file_whole(file_id=resp["file_id"], is_results_file=is_results_file)
        #self.assertEqual(len(test_res_whole["counts"]), len(test_vec_int))
        #self.assertEqual(len(test_res_whole["energies"]), test_vec.shape[0])
        #self.assertEqual(len(test_res_whole["samples"]), num_samples)
        #self.assertEqual(len(test_res_whole["samples"][0]), num_nodes)

        del_resp = self.q1.delete_file(resp["file_id"])
        assert del_resp["num_deleted"] == 1


@pytest.mark.offline
class TestJobsApiWithRequestMocks(unittest.TestCase):
    """Jobs-API-related test suite that can be run without backend."""

    def setUp(self):
        self.qci_client = QciClient(
            api_token="test_api_token", url="test_url", set_bearer_token_on_init=False
        )
        self.job_id = "63b717a22da68618ec444eac"
        self.job_type = "sample-hamiltonian"
        self.file_id = "73b717a22da68618ec444eab"

        # Bad GET jobs response.
        self.get_response_bad = requests.Response()
        self.get_response_bad.status_code = 404

    def test_jobs_url(self) -> None:
        """Test getting jobs URL."""
        self.assertEqual(self.qci_client.jobs_url, "test_url/optimization/v1/jobs")

    def test_get_job_id_url(self) -> None:
        """Test getting jobs URL for given job ID."""
        self.assertEqual(
            self.qci_client.get_job_id_url(self.job_id), f"test_url/optimization/v1/jobs/{self.job_id}"
        )

    def test_get_job_statuses_url(self) -> None:
        """Test getting jobs-statuses URL for given job ID."""
        self.assertEqual(
            self.qci_client.get_job_statuses_url(self.job_id),
            f"test_url/optimization/v1/jobs/{self.job_id}",
        )

    def test_files_url(self) -> None:
        """Test getting files URL."""
        self.assertEqual(self.qci_client.files_url, "test_url/optimization/v1/files")

    def test_get_file_id_url(self) -> None:
        """Test getting files URL for given file ID."""
        self.assertEqual(
            self.qci_client.get_file_id_url(self.file_id),
            f"test_url/optimization/v1/files/{self.file_id}",
        )

    def test_get_job_type_from_job_id(self) -> None:
        """Test getting job type from job ID."""
        # GET short jobs response.
        get_jobs_response_short = requests.Response()
        get_jobs_response_short.status_code = 200
        get_jobs_response_short._content = (  # pylint: disable=protected-access
            b"""{
    "job_id": "%b",
    "status": "COMPLETED",
    "type": "sample_hamiltonian",
    "organization_id": "6edf5db3def87d53b6bf375b",
    "username": "test_user"
}"""
            % self.job_id.encode()
        )

        self.qci_client.session.request = unittest.mock.MagicMock(
            return_value=get_jobs_response_short
        )
        self.assertEqual(
            self.qci_client.get_job_type_from_job_id(self.job_id), self.job_type
        )

        self.qci_client.session.request = unittest.mock.MagicMock(
            return_value=self.get_response_bad
        )
        with self.assertRaises(AssertionError):
            self.qci_client.get_job_type_from_job_id(self.job_id)



"""
QciClient
Utility class for user interactions with QCI API
"""

from dataclasses import dataclass
from datetime import datetime
import gzip
from io import BytesIO
import json
from posixpath import join
import sys
import time
from typing import ClassVar, Optional

from requests.adapters import HTTPAdapter, Retry
from requests_futures.sessions import FuturesSession
import concurrent.futures

from qci_client.base import BaseApi, BACKOFF_FACTOR, RETRY_TOTAL, STATUS_FORCELIST
from qci_client import utilities
from qci_client.data_converter import data_to_json

VERIFY_DEFAULT = True
TIMEOUT_DEFAULT: Optional[float] = 2 * 60.0  # seconds, or None for infinite.
COMPRESS_DEFAULT = False
MAX_WORKERS = 8

class JobStatus:  # pylint: disable=too-few-public-methods
    """Allowed jobs statuses."""

    QUEUED = "QUEUED"
    SUBMITTED = "SUBMITTED"
    RUNNING = "RUNNING"
    COMPLETED = "COMPLETED"
    ERRORED = "ERRORED"
    CANCELLED = "CANCELLED"

FINAL_STATUSES = frozenset(["completed", "errored", "cancelled"])

@dataclass
class QciClient(BaseApi):  # pylint: disable=too-many-public-methods
    """
    Provides requests for QCIs public API as well as utility functions for creating requests and processing entire jobs.
    :param max_workers: int, number of threads for concurrent file download calls
    :param files: url path fragment to specify files API endpoint
    :param jobs: url path fragment to specify jobs API endpoint
    :param _supported_job_types: list of job_types accepted by jobs endpoint
    """
    max_workers: int = 8
    files: str = "optimization/v1/files"
    jobs: str = "optimization/v1/jobs"
    # legacy: bool = False
    compress: bool = COMPRESS_DEFAULT
    verify: bool = VERIFY_DEFAULT
    # _supported_job_types: List[str] = None
    _supported_job_types: ClassVar[frozenset] = frozenset(
        [
            "sample-qubo",
            # "bipartite-community-detection",
            # "unipartite-community-detection",
            "graph-partitioning",
            "sample-constraint",
            "sample-hamiltonian",
            "sample-hamiltonian-ising",
        ]
    )

    @property
    def jobs_url(self):
        """Get jobs URL."""
        return join(self.url, self.jobs)

    def get_job_url(self) -> str:
        """Get job URL with job type."""
        return self.jobs_url

    def get_job_id_url(self, job_id: str) -> str:
        """Get job URL with job ID."""
        return join(self.jobs_url, job_id)

    def get_job_statuses_url(self, job_id: str) -> str:
        """Get job status using job ID."""
        return self.get_job_id_url(job_id)

    def get_job_allocations_url(self) -> str:
        """Get job allocations"""
        return join(self.jobs_url, "allocations")

    @property
    def files_url(self):
        """Get files URL."""
        return join(self.url, self.files)

    def get_file_id_url(self, file_id: str) -> str:
        """Get file URL with file ID."""
        return join(self.files_url, file_id)

    def get_file_contents_url(self, file_id: str, part_num: int) -> str:
        """Get file contents URL with file ID and file part number."""
        return join(self.get_file_id_url(file_id), "contents", str(part_num))

    @property
    def headers_without_authorization(self) -> dict:
        """HTTP headers without bearer token."""
        headers = {
            "Content-Type": "application/json",
            # Simple, sessionless requests, so close connection proactively.
            "Connection": "close",
        }

        if self.timeout is not None:
            # Tell server when client will stop waiting for response.
            headers["X-Request-Timeout-Nano"] = str(int(10**9 * self.timeout))

        return headers

    @BaseApi.refresh_token
    def upload_file(  # pylint: disable=too-many-branches, too-many-locals, too-many-statements
        self,
        file: dict,
    ) -> dict:
        """
        Upload file (metadata and then parts concurrently). Returns dict with file ID.
        """
        # Use session with maintained connection and multipart concurrency for
        # efficiency.
        file_type = utilities.get_file_type(file=file) 
        data = data_to_json(
            data=file
        )

        #data = data_to_json(
        #   data=data['file_config'][file_type]['data'], file_type=file_type, file_name=file_name, debug=self.debug 
        #)
        #if file_type is not None:
        #    if not isinstance(data['file_config'][file_type]['data'], dict):
        #        if self.debug:
        #            print(f"Converting data for file_type={file_type} to JSON...")
        #            print(f"\tElapsed time = {time.perf_counter() - start_time_s} s.")
        # 
        #        data = data_to_json(
        #            data=data['file_config'][file_type]['data'], file_type=file_type, file_name=file_name, debug=self.debug
        #        )


        with FuturesSession(max_workers=self.max_workers) as session:
            session.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=RETRY_TOTAL,
                        backoff_factor=BACKOFF_FACTOR,
                        status_forcelist=STATUS_FORCELIST,
                    )
                ),
            )

            post_response_future = session.post(
                self.files_url,
                headers=self.headers_without_connection_close,
                timeout=self.timeout,
                verify=self.verify,
                json=utilities.get_post_request_body(file=data),
            )

            for response_future in concurrent.futures.as_completed(
                [post_response_future], self.timeout
            ):
                response = response_future.result()
                response.raise_for_status()

            file_id = response.json()["file_id"]
            file_part_generator = utilities.file_part_generator(
                file=data, compress=self.compress
            )
            patch_response_futures = []

            if self.compress:
                for part_body, part_number in file_part_generator:
                    patch_response_futures.append(
                        session.patch(
                            join(self.files_url, f"{file_id}/contents/{part_number}"),
                            headers=self.headers_without_connection_close,  # pylint: disable=line-too-long
                            timeout=self.timeout,
                            verify=self.verify,
                            data=utilities.zip_payload(
                                payload=utilities.get_patch_request_body(file=part_body)
                            ),
                        )
                    )
            else:
                for part_body, part_number in file_part_generator:
                    patch_response_futures.append(
                        session.patch(
                            join(self.files_url, f"{file_id}/contents/{part_number}"),
                            headers=self.headers_without_connection_close,  # pylint: disable=line-too-long
                            timeout=self.timeout,
                            verify=self.verify,
                            json=utilities.get_patch_request_body(file=part_body),
                        )
                    )

            # Due to timeout in underlying PATCH, this should not hang despite no
            # timeout.
            for response_future in concurrent.futures.as_completed(
                patch_response_futures
            ):
                response = response_future.result()
                response.raise_for_status()

        return {"file_id": file_id}

    @BaseApi.refresh_token
    def download_file(self, *, file_id: str) -> dict:
        """Download file (metadata and then parts concurrently)."""
        # Use session with maintained connection and multipart concurrency for
        # efficiency.
        with FuturesSession(max_workers=self.max_workers) as session:
            session.mount(
                "https://",
                HTTPAdapter(
                    max_retries=Retry(
                        total=RETRY_TOTAL,
                        backoff_factor=BACKOFF_FACTOR,
                        status_forcelist=STATUS_FORCELIST,
                    )
                ),
            )

            get_response_future = session.get(
                #urljoin(self.files_url, file_id),
                join(self.files_url, file_id),
                headers=self.headers_without_connection_close,
                timeout=self.timeout,
                #verify=self.verify,
                verify=self.verify,
            )

            for response_future in concurrent.futures.as_completed(
                [get_response_future], self.timeout
            ):
                response = response_future.result()
                response.raise_for_status()

            # File metadata is base for returned fully assembled file.
            file = {**response.json()}

            # Remove metadata fields that are not well-defined for fully assembled file.
            file.pop("last_accessed_rfc3339")
            file.pop("upload_date_rfc3339")

            get_response_futures = [
                session.get(
                    #urljoin(self.files_url, f"{file_id}/contents/{part_number}"),
                    join(self.files_url, f"{file_id}/contents/{part_number}"),
                    headers=self.headers_without_connection_close,
                    timeout=self.timeout,
                    verify=self.verify,
                )
                for part_number in range(1, file["num_parts"] + 1)
            ]

            # Due to timeout in underlying GET, this should not hang despite no timeout.
            for response_future in concurrent.futures.as_completed(
                get_response_futures
            ):
                response = response_future.result()
                response.raise_for_status()

            # Unpack in order.
            for response_future in get_response_futures:
                file_part = response_future.result().json()
                # Append to all array fields.
                for file_type, file_type_config in file_part["file_config"].items():
                    if file_type not in file["file_config"]:
                        file["file_config"][file_type] = {}

                    for key, value in file_type_config.items():
                        if key not in file["file_config"][file_type]:
                            file["file_config"][file_type][key] = []

                        file["file_config"][file_type][key] += value

        return file

    @BaseApi.refresh_token
    def submit_job(self, job_body: dict, job_type: str) -> dict:
        """
        Submit a job via a request to QCI public API.

        Args:
            job_body: formatted json body that includes all parameters for the job
            job_type: one of the _supported_job_types

        Returns:
            Response from POST call to API
        """
        self.validate_job_type(job_type=job_type)
        response = self.session.request(
            "POST",
            self.get_job_url(),
            json=job_body,
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )
        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def get_job_status(self, job_id: str) -> dict:
        """
        Get the status of a job by its ID.

        Args:
            job_id: ID of job

        Returns:
            Response from GET call to API
        """
        response = self.session.request(
            "GET",
            self.get_job_statuses_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )

        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def get_job_response(self, job_id: str, job_type: str) -> dict:
        """
        Get a response for a job by id and type, which may/may not be finished.

        :param job_id: ID of job
        :param job_type: type of job, one of []

        :return dict: loaded json file
        """
        self.validate_job_type(job_type=job_type)
        response = self.session.request(
            "GET",
            self.get_job_id_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )

        self._check_response_error(response=response)
        return response.json()

    def validate_job_type(self, job_type: str) -> None:
        """
        Checks if a provided job type is a supported job type.

        Args:
            job_type: a job type to validate

        Returns:
            None

        Raises AssertionError if job_type is not one of the _supported_job_types.
        """
        assert (
            job_type in self._supported_job_types
        ), f"Provided job_type '{job_type}' is not one of {self._supported_job_types}"

    def build_job_body(  # pylint: disable=too-many-arguments
        self,
        job_type: str,
        relaxation_schedule: Optional[int] = 2,
        qubo_file_id: Optional[str] = None,
        graph_file_id: Optional[str] = None,
        hamiltonian_file_id: Optional[str] = None,
        objective_file_id: Optional[str] = None,
        constraints_file_id: Optional[str] = None,
        rhs_file_id: Optional[str] = None,
        job_params: Optional[dict] = None,
        job_name: str = "job_0",
        job_tags: Optional[list] = None,
        solution_precision: Optional[float] = 1,
    ) -> dict:
        """
        Constructs body for job submission requests
        :param job_type: one of _supported_job_types
        :param qubo_file_id: file id from files API for uploaded qubo
        :param graph_file_id: file id from files API for uploaded graph
        :param hamiltonian_file_id: file id from files API for uploaded hamiltonian
        :param objective_file_id: file id from files API for uploaded objective
        :param constraints_file_id: file id from files API for uploaded constraints
        :param rhs_file_id: file id from files API for uploaded rhs
        :param job_params: dict of additional params to be passed to job submission in "params" key
        :param job_name: user specified name for job submission
        :param job_tags: user specified labels for classifying and filtering user jobs after submission
        :param relaxation_schedule: user specified relaxation schedule for problem execution
        :note: Need to add validation for job parameters
        """
        if job_params is None:
            job_params = {}

        if job_tags is None:
            job_tags = []

        self.validate_job_type(job_type=job_type)
        assert (
            sum(
                fid is not None
                for fid in [
                    qubo_file_id,
                    graph_file_id,
                    hamiltonian_file_id,
                    objective_file_id,
                ]
            )
            == 1
        ), "Only one of qubo_file_id, hamiltonian_file_id, objective_file_id, or graph_file_id can be specified"

        # TODO: remove this in the future for now map dirac- to eqc but warn of deprecation
        if "eqc" in job_params["sampler_type"]:
            print("WARNING: " + job_params['sampler_type'] + " will be a deprecated sampler type dirac-(1-3) will be the supported sampler types in the future")
            job_params["sampler_type"] = job_params["sampler_type"].replace("eqc", "dirac-")

        # n_samples num_samples
        if "n_samples" in job_params:
            print("WARNING: the key n_samples will be a deprecated parameter in the future, nsamples will be the supported parameter.")
            job_params["nsamples"] = job_params.pop("n_samples")
        elif "num_samples" in job_params:
            print("WARNING: the key num_samples will be a deprecated parameter in the future, nsamples will be the supported parameter.")
            job_params["nsamples"] = job_params.pop("num_samples")
        elif "num_solutions" in job_params:
            print("WARNING: the key num_solutions will be a deprecated parameter in the future, nsamples will be the supported parameter.")
            job_params["nsamples"] = job_params.pop("num_solutions")
        else:
            pass

        config_value = ""
        job_body = {"job_name": job_name, "job_tags": job_tags}
        device_settings = {}
        if "relaxation_schedule" in job_params:
            device_settings["relaxation_schedule"] = job_params["relaxation_schedule"]
        if job_type == 'sample-qubo':
            config_value = 'quadratic_unconstrained_binary_optimization'
        elif job_type == 'sample-hamiltonian':
            if "solution_precision" in job_params and job_params["solution_precision"] == 1:
                config_value = 'normalized_qudit_hamiltonian_optimization_integer'
                job_body.update(
                    {"solution_precision": solution_precision,}
                )
            elif "solution_precision" in job_params and job_params["solution_precision"] != 1:
                config_value = 'normalized_qudit_hamiltonian_optimization_continuous'
                job_body.update(
                    {"solution_precision": solution_precision,}
                )
            else:
                config_value = 'normalized_qudit_hamiltonian_optimization_integer'
            device_settings["sum_constraint"] = job_params["sum_constraint"]
        elif job_type == 'sample-hamiltonian-ising':
            config_value = 'ising_hamiltonian_optimization'
        problem_config = {"problem_config": {config_value: job_body}, "device_config": {job_params['sampler_type']: device_settings}}
        job_submission = {"job_submission": problem_config}

        if job_type == "sample-constraint":
            assert None not in [
                objective_file_id,
                constraints_file_id,
                #rhs_file_id,
            ], "objective_file_id, constraints_file_id, and rhs_file_id must all be specified for job_type='sample-constraint'"
            job_body.update(
                {
                    "objective_file_id": objective_file_id,
                    "constraints_file_id": constraints_file_id,
                    "rhs_file_id": rhs_file_id,
                    "num_samples": job_params["nsamples"],
                }
            )
            config_value = 'quadratic_linearly_constrained_binary_optimization'
        else:
            assert all(
                fid is None
                for fid in [constraints_file_id, rhs_file_id, objective_file_id]
            ), "objective_file_id, constraints_file_id, and rhs_file_id are not available for selected job_type"
            if job_type == "sample-qubo":
                assert (
                    qubo_file_id is not None
                ), "qubo_file_id must be specified for job_type='sample-qubo'"
                job_body["qubo_file_id"] = qubo_file_id
                job_body["num_samples"] = job_params['nsamples']
            elif job_type == "sample-hamiltonian":
                assert (
                    hamiltonian_file_id is not None
                ), "hamiltonian_file_id must be specified for job_type='sample-hamiltonian'"
                job_body["hamiltonian_file_id"] = hamiltonian_file_id
                #if job_params['sampler_type'] == 'dirac-2' or job_params['sampler_type'] == 'dirac-3':
                #    job_body["solution_type"] = job_params["solution_type"]
                #else:
                #    print(f"{job_type} is not supported for dirac-1 or csample. Consider using dirac-2 or dirac-3 for this job type.")
                #    sys.exit(1)
                job_body["num_samples"] = job_params['nsamples']
            elif job_type == "sample-hamiltonian-ising":
                if job_params['sampler_type'] == 'dirac-2' or job_params['sampler_type'] == 'dirac-3':
                    print(f"{job_type} not supported on dirac-2 and dirac-3. Consider using job_type sample-hamiltonian for dirac-2 and dirac-3.")
                    sys.exit(1)
                elif job_params['sampler_type'] == 'csample':
                    print(f"{job_type} not currently supported on csample. Consider running on dirac-1.")
                    sys.exit(1)
                assert (
                    hamiltonian_file_id is not None
                ), "hamiltonian_file_id must be specified for job_type='sample-hamiltonian'"
                job_body["hamiltonian_file_id"] = hamiltonian_file_id
                job_body["num_samples"] = job_params['nsamples']
            else:
                # if add more jobs that are non graph would need to change this else
                assert (
                    graph_file_id is not None
                ), "graph_file_id must be specified for the given job_type"
                job_body["graph_file_id"] = graph_file_id
        problem_config =  {"problem_config": {config_value: job_body}, "device_config": {job_params['sampler_type']: device_settings}}
        job_submission = {"job_submission": problem_config}

        # list of job_params vary by job_type
        # could add a validate params function here
        return job_submission

    def print_job_log(self, message: str) -> None:
        """
        Formats a messages for updating user with a time stamp appended
        :param message: a string to be passed in print statement
        """
        print(f"{message}: {datetime.now().strftime('%Y/%m/%d %H:%M:%S')}")

    # TODO: add provider and device names args (optional)

    def process_job(self, job_type: str, job_body: dict, wait: bool = True) -> dict:
        """
        :param job_type: the type of job being processed must be one of _supported_job_types
        :param job_body: formatted json dict for body of job submission request
        :param wait: bool indicating whether or not user wants to wait for job to complete

        :return:
            if wait is True, then dict with job_info response and results file
                (results is None if ERRORED or CANCELLED)
            if wait is False, then response dict from submitted job, which includes job
                ID for subsequent retrieval
        :note: what else do we want to return with the results? response_id, obviously job_id
        """
        allocations_response = self.session.request(
            "GET",
            self.get_job_allocations_url(),
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )
        dirac_allocation = allocations_response.json()['allocations']['dirac']['seconds']
        print(f"Dirac allocation balance = {dirac_allocation}")
        self.validate_job_type(job_type=job_type)
        submit_response = self.submit_job(job_body=job_body, job_type=job_type)
        job_id = submit_response["job_id"]
        self.print_job_log(message=f"Job submitted job_id='{job_id}'-")
        curr_status = None
        if wait:
            while curr_status not in FINAL_STATUSES:
                time.sleep(1)
                status_response = self.get_job_status(job_id=job_id)
                iter_status = status_response["job_status"]
                iter_status = (list(iter_status.keys()))
                iter_status = iter_status[-1]
                iter_status = iter_status.replace('_at_rfc3339nano', '')
                if iter_status != curr_status:
                    self.print_job_log(message=iter_status)
                    curr_status = iter_status

            job_response = self.get_job_response(job_id=job_id, job_type=job_type)
            if curr_status.upper() in [JobStatus.CANCELLED, JobStatus.ERRORED]:
                results = None
                allocations_response = self.session.request(
                    "GET",
                    self.get_job_allocations_url(),
                    headers=self.headers,
                    timeout=self.timeout,
                    verify=self.verify,
                )
                dirac_allocation = allocations_response.json()['allocations']['dirac']['seconds']
                print(f"Dirac allocation balance = {dirac_allocation}")
            else:
                results_fid = job_response["job_result"]["file_id"]
                details = {}
                details['status'] = iter_status
                job_response['details'] = details
                results = self.download_file(file_id=results_fid)
                allocations_response = self.session.request(
                    "GET",
                    self.get_job_allocations_url(),
                    headers=self.headers,
                    timeout=self.timeout,
                    verify=self.verify,
                )
                dirac_allocation = allocations_response.json()['allocations']['dirac']['seconds']
                print(f"Dirac allocation balance = {dirac_allocation}")
            return {"job_info": job_response, "results": results}

        return submit_response

    @BaseApi.refresh_token
    def list_files(self, username: Optional[str] = None) -> dict:
        """
        :param username: Optional str - username (to search for files owned by the named user)
            mostly useful when run by users with administrator privileges (such as QCI users) who can see all files.
            When called by an administrator, the username parameter is used to restrict the list files returned
            to be only the files owned by the user specified in the username parameter.
            When run by non-privileged users, this parameter is truly optional because non-privileged users
            will only ever see lists of files that they created.

        :return: dict containing list of files
        """
        if username:
            querystring = {"regname": "username", "regvalue": username}

            response = self.session.request(
                "GET",
                self.files_url,
                headers=self.headers,
                params=querystring,
                timeout=self.timeout,
                verify=self.verify,
            )
        else:
            response = self.session.request(
                "GET",
                self.files_url,
                headers=self.headers,
                timeout=self.timeout,
                verify=self.verify,
            )

        self._check_response_error(response=response)
        return response.json()

    @BaseApi.refresh_token
    def delete_file(self, file_id: str) -> dict:
        """
        :param file_id: str - file_id of file to be deleted

        :return: dict containing information about file deleted (or error)
        """

        if self.debug:
            print(f"Deleting file with ID {file_id}...")

        start_time_s = time.perf_counter()

        response = self.session.request(
            "DELETE",
            self.get_file_id_url(file_id),
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )

        stop_time_s = time.perf_counter()

        if self.debug:
            print(f"Deleting file with ID {file_id}...done.")
            print(f"\tElapsed time: {stop_time_s - start_time_s} s.")

        self._check_response_error(response=response)

        return response.json()

    @BaseApi.refresh_token
    def zip_payload(self, payload: str) -> bytes:
        """
        :param payload: str - json contents of file to be zipped

        return: zipped request_body
        """
        out = BytesIO()
        with gzip.GzipFile(fileobj=out, mode="w", compresslevel=6) as file:
            file.write(json.dumps(payload).encode("utf-8"))
        request_body = out.getvalue()
        out.close()
        return request_body

    @BaseApi.refresh_token
    def merge_two_dicts(self, x, y):  # pylint: disable=invalid-name
        """Given two dictionaries, merge them into a new dict as a shallow copy."""
        z = x.copy()  # pylint: disable=invalid-name
        z.update(y)
        return z

    def get_job_type_from_job_id(self, job_id: str) -> str:
        """
        Get job type from job ID.

        Args:
            job_id: ID of the job

        Returns:
            Type of the job
        """
        response_job_metadata_short = self.session.request(
            "GET",
            self.get_job_id_url(job_id),
            headers=self.headers,
            timeout=self.timeout,
            verify=self.verify,
        )
        self._check_response_error(response=response_job_metadata_short)

        return response_job_metadata_short.json()["type"].replace("_", "-")

    @BaseApi.refresh_token
    def _get_results_file_contents_url(self, file_id: str, part_num: int) -> str:
        """
        Adheres to this route: 'results/:id/contents/:part-num'

        :param file_id: str, file_id
        :param part_num: int, file chunk part number. File part_num starts counting at 1.
        :return: str, the joined url
        """
        return f"{self.files_url}/results/{file_id}/contents/{part_num}"




"""Functions for data conversion."""

import json
from math import floor
import sys
import time
from typing import Generator, Optional, Union

import networkx as nx
from networkx.readwrite import json_graph
import numpy as np
import scipy.sparse as sp

from qci_client import utilities

# We want to limit the memory size of each uploaded chunk to be safely below the max of 15 * MebiByte (~15MB).
# See https://git.qci-dev.com/qci-dev/qphoton-files-api/-/blob/main/service/files.go#L32.
MEMORY_MAX: int = 8 * 1000000  # 8MB


def load_json_file(file_name: str) -> dict:
    """
    Load a utf-8-encoded json file into a dictionary.

    :param file_name: name of the JSON file to load

    :return dict: loaded json file
    """
    with open(file_name, "r", encoding="utf-8") as file:
        return json.loads(file.read())


def get_size(obj, seen=None) -> int:
    """
    Recursively finds size of objects

    :param obj: data object to recursively compute size of
    :param seen: takes a set and is used in the recursive step only to record whether an object has been counted yet.

    :return int:
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum(get_size(v, seen) for v in obj.values())
        size += sum(get_size(k, seen) for k in obj.keys())
    elif hasattr(obj, "__dict__"):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
        size += sum(get_size(i, seen) for i in obj)
    return size


def _get_soln_size(soln):
    # Check whether first entry is a graph node/class assignment, eg., {'id': 4, 'class': 2}
    if isinstance(soln[0], dict):
        return get_size(soln)

    return sys.getsizeof(soln[0]) * len(soln)


def compute_results_step_len(data: Union[np.ndarray, list]) -> int:
    """
    Compute the step length for "chunking" the providd data.

    Args:
        data: An numpy array or list of data

    Returns:
        The step length for "chunking" the data
    """
    # total mem size of soln vector
    soln_mem = _get_soln_size(data)
    # num_vars * step_len < 30k => step_len < 30k/num_vars
    chunk_ratio = MEMORY_MAX / soln_mem
    step_len = floor(chunk_ratio) if chunk_ratio >= 1 else 1
    return step_len

def _data_generator(data: dict, step_length: int, legacy: bool) -> Generator:
    df = data["data"]  # pylint: disable=invalid-name

    # df may be empty, so use max(1, len(df)).
    if legacy:
        for part_num, i in enumerate(range(0, max(1, len(df)), step_length)):
            if part_num == 0:
                chunk = {key: val for key, val in data.items() if key not in ["data"]}
                chunk["data"] = df[i : i + step_length]
            else:
                chunk = {"file_type": data["file_type"], "data": df[i : i + step_length]}
            yield chunk, part_num

    else:
        for part_num, i in enumerate(range(1, max(1, len(df)), step_length)):
            if part_num == 0:
                chunk = {key: val for key, val in data.items() if key not in ["data"]}
                chunk["data"] = df[i : i + step_length]
            else:
                chunk = {"file_type": data["file_type"], "data": df[i : i + step_length]}
            #print("PART NUM")
            #print(part_num)
            yield chunk, part_num + 1

def _graph_generator(data: dict, step_length: int) -> Generator:
    num_links = len(data["links"])
    num_nodes = len(data["nodes"])

    # links and nodes may both be empty, so use max(1, num_links, num_nodes).
    for part_num, i in enumerate(range(0, max(1, num_links, num_nodes), step_length)):
        if part_num == 0:
            chunk = {
                key: val for key, val in data.items() if key not in ["links", "nodes"]
            }
            chunk["links"] = data["links"][i : i + step_length]
            chunk["nodes"] = data["nodes"][i : i + step_length]
        else:
            # patch requires 'file_type' and 'graph', as well as node and link chunks
            chunk = {
                "file_type": data["file_type"],
                "graph": data["graph"],
                # Lengths below are unequal, meaning that one of these will return an empty list if indices DNE
                "links": data["links"][i : i + step_length],
                "nodes": data["nodes"][i : i + step_length],
            }
        yield chunk, part_num


def _results_generator(data: dict, step_length: int) -> Generator:
    file_type = data["file_type"]
    #
    # GRAPH PARTITIONING
    #
    if file_type.split("_")[-1] == "partitioning":
        for part_num, i in enumerate(range(0, len(data["samples"]), step_length)):
            chunk = {"file_type": data["file_type"], "file_name": data["file_name"]}
            if part_num == 0:
                chunk.update(
                    {
                        key: val
                        for key, val in data.items()
                        if key
                        not in [
                            "samples",
                            "energies",
                            "counts",
                            "balance",
                            "cut_size",
                            "is_feasible",
                        ]
                    }
                )
            chunk.update(
                {
                    "samples": data["samples"][i : i + step_length],
                    "energies": data["energies"][i : i + step_length],
                    "counts": data["counts"][i : i + step_length],
                    "balance": data["balance"][i : i + step_length],
                    "cut_size": data["cut_size"][i : i + step_length],
                    "is_feasible": data["is_feasible"][i : i + step_length],
                }
            )
            yield chunk, part_num
    #
    # (UNI AND BI)-COMMUNITY DETECTION
    #
    elif file_type.split("_")[-1] == "detection":
        for part_num, i in enumerate(range(0, len(data["samples"]), step_length)):
            chunk = {"file_type": data["file_type"]}
            if part_num == 0:
                # TODO: split up counts and energies too. They are also expected by API in each upload
                chunk.update(
                    {
                        key: val
                        for key, val in data.items()
                        if key
                        not in [
                            "samples",
                            "energies",
                            "counts",
                            "modularity",
                            "is_feasible",
                        ]
                    }
                )
            chunk.update(
                {
                    "samples": data["samples"][i : i + step_length],
                    "energies": data["energies"][i : i + step_length],
                    "counts": data["counts"][i : i + step_length],
                    "modularity": data["modularity"][i : i + step_length],
                    "is_feasible": data["is_feasible"][i : i + step_length],
                }
            )
            yield chunk, part_num
    #
    # NON-GRAPH RESULTS DATA
    #
    else:
        # Number of samples should align with len(energies) and len(counts)
        for part_num, i in enumerate(range(0, len(data["samples"]), step_length)):
            chunk = {"file_type": data["file_type"]}
            if part_num == 0:
                chunk.update(
                    {
                        key: val
                        for key, val in data.items()
                        if key not in ["samples", "energies", "counts"]
                    }
                )
            chunk.update(
                {
                    "samples": data["samples"][i : i + step_length],
                    "energies": data["energies"][i : i + step_length],
                    "counts": data["counts"][i : i + step_length],
                }
            )
            yield chunk, part_num


def build_graph_data_from_get_file(file_parts: Generator, file_type: str) -> dict:
    """
    Return graph data built from a given graph file generator and type.

    Args:
        file_parts: File-part generator
        file_type: Type of the file

    Returns:
        Dictionary of graph data
    """
    # Each case needs to extract metadata from the first element off the heap
    first_part = next(file_parts)
    if "graph" == file_type:
        #
        # REBUILD A LARGE GRAPH FILE
        #
        file_data = {
            key: val for key, val in first_part.items() if key not in ["links", "nodes"]
        }
        first_part = _munge_node_link_format(first_part)
        file_data["links"] = first_part["links"]
        file_data["nodes"] = first_part["nodes"]
        for part in file_parts:
            scrubbed_data = _munge_node_link_format(part)
            file_data["links"] += scrubbed_data["links"]
            file_data["nodes"] += scrubbed_data["nodes"]
        return file_data

    if file_type.split("_")[-1] == "detection":
        #
        # COMMUNITY DETECTION RESULTS
        #
        file_data = dict(first_part.items())
        for part in file_parts:
            file_data["samples"] += part["samples"]
            file_data["energies"] += part["energies"]
            file_data["counts"] += part["counts"]
            file_data["modularity"] += part["modularity"]
            file_data["is_feasible"] += part["is_feasible"]
    elif file_type.split("_")[-1] == "partitioning":
        #
        # GRAPH PARTITIONING RESULTS
        #
        file_data = dict(first_part.items())
        for part in file_parts:
            file_data["samples"] += part["samples"]
            file_data["energies"] += part["energies"]
            file_data["counts"] += part["counts"]
            file_data["cut_size"] += part["cut_size"]
            file_data["balance"] += part["balance"]
            file_data["is_feasible"] += part["is_feasible"]
    else:
        file_data = dict(first_part.items())
        for part in file_parts:
            file_data["samples"] += part["samples"]
            file_data["energies"] += part["energies"]
            file_data["counts"] += part["counts"]
    # we don't need next_part anymore
    #file_data.pop("next_part")

    return file_data


def _munge_node_link_format(data: dict) -> dict:
    """
    Temporary hack to clean incoming data. Mongo BSON is adding key/value pairs to the node_link format, such as

        'nodes': [[{'Key': 'bipartite', 'Value': 0},
                {'Key': 'id', 'Value': 'Evelyn Jefferson'}],
                [{'Key': 'bipartite', 'Value': 0}, ... ]

        This should look like

        'nodes': [{'bipartite': 0, 'id': 'Evelyn Jefferson'},
                  {'bipartite': 0, 'id': 'Laura Mandeville'},
                  {'bipartite': 0, 'id': 'Theresa Anderson'}, ... ]

    This method just munges the first format in the second format.

    Args:
        data: dict,

    Returns:
        dict, node_link graph
    """
    node_link = {}
    for element in data.keys():
        if element in ("links", "nodes"):
            node_link[element] = []
            # node may exhaust earlier than links, so we only append as needed
            if len(data[element]) > 0:
                for item_list in data[element]:
                    item_props = {item["Key"]: item["Value"] for item in item_list}
                    node_link[element].append(item_props)
    return node_link


def data_to_json(
    data: Union[np.ndarray, sp.spmatrix, nx.Graph, list],
    debug: bool = False,
) -> dict:
    """
    Converts  data input into JSON string that can be passed to Qatalyst REST API
    :param data: object to be converted to JSON
       * currently constraint_penalties require one value for each constraint
    :param file_type: one of ["graph", "qubo", "objective", "constraints", "constraint_penalties", "rhs", "hamiltonian"]
    :param file_name: Optional user specified name for file to be uploaded
    :param debug: Optional, if set to True, enables debug output (default = False for no debug output)

    :return: string (JSON format)
    :note: could add support for matrices stored as lists
    """
    supported_file_types = [
        "graph",
        "qubo",
        "hamiltonian",
        "rhs",
        "objective",
        "constraints",
        "constraint_penalties",
    ]

    result_file_types = [
        "job_results_sample_qubo",
        "job_results_sample_hamiltonian",
        "job_results_sample_constraint",
        "job_results_unipartite_community_detection",
        "job_results_bipartite_community_detection",
        "job_results_graph_partitioning",
    ]

    supported_file_types += result_file_types

    file_type = utilities.get_file_type(file=data).value
    start = time.perf_counter()
    assert (
        file_type in supported_file_types
    ), f"Unsupported file type input specify one of {supported_file_types}"
    file_name = data.get("file_name", f"{file_type}.json")
    data = data['file_config'][file_type]['data']
    file_config = {}
    file_body = {"file_name": file_name, "file_config": {file_type: file_config}}
    if file_type == "graph":
        assert isinstance(
            data, nx.Graph
        ), "'graph' file_type must be type networkx.Graph"
        data = json_graph.node_link_data(data)
    else:
        assert not isinstance(data, nx.Graph), (
            "file_types ['rhs', 'objective', 'qubo', 'constraints', 'constraint_penalties', 'hamiltonian'] "
            "do not support networkx.Graph type"
        )
        if file_type in ["rhs", "constraint_penalties"]:
            if isinstance(data, sp.spmatrix):
                data = data.toarray()
            if isinstance(data, np.ndarray):
                # deal with case where shape (1,n) or (n,1)
                data = data.flatten()
                data = data.tolist()
            file_config["num_constraints"] = len(data)
        else:
            assert isinstance(data, (sp.spmatrix, np.ndarray)), (
                "file_types = ['qubo', 'objective', 'constraints', 'hamiltonian'] only support types np.ndarray "
                "and scipy.sparse.spmatrix"
            )
            data = sp.coo_matrix(data)
            rows, cols = data.shape

            data_ls = []
            for i, j, val in zip(
                data.row.tolist(), data.col.tolist(), data.data.tolist()
            ):
                data_ls.append({"i": i, "j": j, "val": val})
            # must reassign to var that will be used in final update to dict
            data = data_ls
            if file_type == "constraints":
                file_config.update({"num_constraints": rows, "num_variables": cols})
            else:
                file_config["num_variables"] = rows
        # update both rhs and all objective and constraint data key
        data = {"data": data}
    # use update because graph data has more than one entry
    file_config.update(data)
    if debug:
        print(f"Time to convert data to json: {time.perf_counter()-start} s.")

    return file_body
